from __future__ import annotations
from dataclasses import dataclass, asdict
import os
import json
from enum import Enum
from pathlib import Path
from typing import Optional, Any
from PyQt5.QtCore import QObject, pyqtSignal

from .util import is_macos, is_windows, user_data_dir, client_logger as log
from .util import encode_json, read_json_with_comments


class ServerMode(Enum):
    undefined = -1
    managed = 0
    external = 1
    cloud = 2


class ServerBackend(Enum):
    cpu = ("CPUで実行", True)
    cuda = ("CUDAを使用(NVIDIA GPU)", not is_macos)
    mps = ("MPSを使用(Mac)", is_macos)
    directml = ("DirectMLを使用(GPU)", is_windows)

    @staticmethod
    def supported():
        return [b for b in ServerBackend if b.value[1]]

    @staticmethod
    def default():
        if is_macos:
            return ServerBackend.mps
        else:
            return ServerBackend.cuda


class ApplyBehavior(Enum):
    replace = 0
    layer = 1
    layer_group = 2
    layer_hide_below = 3
    transparency_mask = 4


class PerformancePreset(Enum):
    auto = "自動"
    cpu = "CPU"
    low = "GPU 低VRAM (6GB～)"
    medium = "GPU 中VRAM (6GB～12GB)"
    high = "GPU 高VRAM (12GB～)"
    cloud = "クラウド"
    custom = "カスタム"


@dataclass
class PerformanceSettings:
    batch_size: int = 4
    resolution_multiplier: float = 1.0
    max_pixel_count: int = 6


class Setting:
    def __init__(self, name: str, default, desc="", help="", items=None):
        self.name = name
        self.desc = desc
        self.default = default
        self.help = help
        self.items = items

    def str_to_enum(self, s: str):
        assert isinstance(self.default, Enum)
        EnumType = type(self.default)
        try:
            return EnumType[s]
        except KeyError:
            return self.default


class Settings(QObject):
    default_path = user_data_dir / "settings.json"

    server_mode: ServerMode
    _server_mode = Setting(
        "サーバー管理",
        ServerMode.undefined,
        "画像を生成するために、プラグインはComfyUIサーバーに接続します。",
    )

    access_token: str
    _access_token = Setting("Cloud Access Token", "")

    server_path: str
    _server_path = Setting(
        "サーバーのパス",
        str(user_data_dir / "server"),
        "ComfyUIがインストールされるディレクトリ。最低限のインストールには、"
        "10GB以上の空きディスク容量が必要です。",
    )

    server_url: str
    _server_url = Setting(
        "サーバーのURL",
        "127.0.0.1:8188",
        "実行中のComfyUIサーバーへ接続するために使用するURL。デフォルトは 127.0.0.1:8188 (ローカル) です。",
    )

    server_backend: ServerBackend
    _server_backend = Setting("サーバーのバックエンド", ServerBackend.default())

    server_arguments: str
    _server_arguments = Setting(
        "サーバーの引数", "", "サーバーに渡される追加のコマンドライン引数"
    )

    selection_grow: int
    _selection_grow = Setting(
        "選択範囲の拡大", 5, "選択領域を指定サイズだけ拡大します"
    )

    selection_feather: int
    _selection_feather = Setting(
        "選択範囲のぼかし", 5, "指定サイズだけ境界線のぼかしを行います"
    )

    selection_padding: int
    _selection_padding = Setting(
        "選択範囲の余白", 7, "選択領域の周囲に最小限の余白を追加します"
    )

    new_seed_after_apply: bool
    _new_seed_after_apply = Setting(
        "ライブの適用後に新しいシード",
        False,
        "ライブモードで結果をキャンバスにコピーした後は、新しいシードを生成",
    )

    prompt_line_count: int
    _prompt_line_count = Setting(
        "プロンプトの行数", 2, "画像を説明するテキストエディタの行数"
    )

    show_negative_prompt: bool
    _show_negative_prompt = Setting(
        "否定的なプロンプト", False, "生成したくない要素を記述するテキストエディタの表示"
    )

    auto_preview: bool
    _auto_preview = Setting(
        "自動プレビュー", True, "生成された最初の結果をキャンバス上に自動プレビュー"
    )

    apply_behavior: ApplyBehavior
    _apply_behavior = Setting(
        "挙動を適用",
        ApplyBehavior.layer_hide_below,
        "結果画像をキャンバスに適用する方法を選択します（生成ワークスペース）",
    )

    apply_behavior_live: ApplyBehavior
    _apply_behavior_live = Setting(
        "挙動を適用(ライブ)",
        ApplyBehavior.replace,
        "ライブモードで結果画像をキャンバスに適用する方法を選択します",
    )

    show_builtin_styles: bool
    _show_builtin_styles = Setting("プリインストールされたスタイルを表示", True)

    history_size: int
    _history_size = Setting(
        "アクティブな履歴のサイズ", 1000, "生成された画像の履歴に使用されるメインメモリー(RAM)のサイズ"
    )

    history_storage: int
    _history_storage = Setting(
        "保存される履歴のサイズ", 20, "生成された画像をディスク上の .kra ファイルに保存するために使用するメモリーサイズ"
    )

    performance_preset: PerformancePreset
    _performance_preset = Setting(
        "パフォーマンスのプリセット",
        PerformancePreset.auto,
        "利用可能なハードウェアに合わせてパフォーマンス設定を構成します。",
    )

    batch_size: int
    _batch_size = Setting(
        "最大バッチサイズ",
        4,
        "一度に複数の画像を生成することで効率を上げます。",
    )

    resolution_multiplier: float
    _resolution_multiplier = Setting(
        "解像度の倍率",
        1.0,
        "生成時の解像度係数。1.0以下の値は高解像度時のパフォーマンスが"
        "向上します。",
    )

    max_pixel_count: int
    _max_pixel_count = Setting(
        "最大画素数",
        6,
        "画像生成の際の最大解像度 (メガピクセル単位) (フルHD ~ 2MP、4k ~ 8MP)。",
    )

    _performance_presets = {
        PerformancePreset.cpu: PerformanceSettings(
            batch_size=1,
            resolution_multiplier=1.0,
            max_pixel_count=2,
        ),
        PerformancePreset.low: PerformanceSettings(
            batch_size=2,
            resolution_multiplier=1.0,
            max_pixel_count=2,
        ),
        PerformancePreset.medium: PerformanceSettings(
            batch_size=4,
            resolution_multiplier=1.0,
            max_pixel_count=6,
        ),
        PerformancePreset.high: PerformanceSettings(
            batch_size=6,
            resolution_multiplier=1.0,
            max_pixel_count=8,
        ),
        PerformancePreset.cloud: PerformanceSettings(
            batch_size=8,
            resolution_multiplier=1.0,
            max_pixel_count=6,
        ),
    }

    debug_dump_workflow: bool
    _debug_dump_workflow = Setting(
        "ワークフローをダンプ", False, "テストとデバッグのために最新のComfyUIプロンプトをログ・フォルダに書き込み"
    )

    # Folder where intermediate images are stored for debug purposes (default: None)
    debug_image_folder = os.environ.get("KRITA_AI_DIFFUSION_DEBUG_IMAGE")

    changed = pyqtSignal(str, object)

    _values: dict[str, Any]

    def __init__(self):
        super().__init__()
        self.restore(init=True)

    def __getattr__(self, name: str):
        if name in self._values:
            return self._values[name]
        return object.__getattribute__(self, name)

    def __setattr__(self, name: str, value):
        if name in self._values:
            self._values[name] = value
            self.changed.emit(name, value)
            if name == "performance_preset":
                self.apply_performance_preset(value)
        else:
            object.__setattr__(self, name, value)

    def restore(self, init=False):
        self.__dict__["_values"] = {
            k[1:]: v.default for k, v in Settings.__dict__.items() if isinstance(v, Setting)
        }
        if not init:
            self.server_mode = ServerMode.managed

    def save(self, path: Optional[Path] = None):
        path = self.default_path or path
        with open(path, "w") as file:
            file.write(json.dumps(self._values, default=encode_json, indent=4))

    def load(self, path: Optional[Path] = None):
        path = self.default_path or path
        self._migrate_legacy_settings(path)
        if not path.exists():
            self.save()  # create new file with defaults
            return

        log.info(f"Loading settings from {path}")
        try:
            contents = read_json_with_comments(path)
            for k, v in contents.items():
                setting = getattr(Settings, f"_{k}", None)
                if setting is not None:
                    if isinstance(setting.default, Enum):
                        self._values[k] = setting.str_to_enum(v)
                    elif isinstance(setting.default, type(v)):
                        self._values[k] = v
                    else:
                        log.error(f"{path}: {v} is not a valid value for '{k}'")
                        self._values[k] = setting.default
        except Exception as e:
            log.error(f"Failed to load settings: {e}")

    def apply_performance_preset(self, preset: PerformancePreset):
        if preset not in [PerformancePreset.custom, PerformancePreset.auto]:
            for k, v in asdict(self._performance_presets[preset]).items():
                self._values[k] = v

    def _migrate_legacy_settings(self, path: Path):
        if path == self.default_path:
            legacy_path = Path(__file__).parent / "settings.json"
            if legacy_path.exists() and not path.exists():
                try:
                    legacy_path.rename(path)
                    log.info(f"Migrated settings from {legacy_path} to {path}")
                except Exception as e:
                    log.warning(f"Failed to migrate settings from {legacy_path} to {path}: {e}")


settings = Settings()
